/**
 * @module viw-webgl-viewer/rendering
 */
import * as THREE from 'three';
import { IRenderer, Scene } from '../../vim-loader/scene';
import { Viewport } from '../viewport';
import { RenderScene } from './renderScene';
import { ViewerMaterials } from '../../vim-loader/materials/viewerMaterials';
import { CSS2DRenderer } from 'three/examples/jsm/renderers/CSS2DRenderer';
import { Camera } from '../camera/camera';
import { RenderingSection } from './renderingSection';
import { Settings } from '../settings/viewerSettings';
/**
 * Manages how vim objects are added and removed from the THREE.Scene to be rendered
 */
export declare class Renderer implements IRenderer {
    /**
     * The THREE WebGL renderer.
     */
    readonly renderer: THREE.WebGLRenderer;
    /**
     * The THREE sample ui renderer
     */
    readonly textRenderer: CSS2DRenderer;
    /**
     * Interface to interact with section box directly without using the gizmo.
     */
    readonly section: RenderingSection;
    /**
     * Determines whether antialias will be applied to rendering or not.
     */
    antialias: boolean;
    private _scene;
    private _viewport;
    private _camera;
    private _composer;
    private _materials;
    private _renderText;
    private _skipAntialias;
    private _needsUpdate;
    private _onSceneUpdate;
    private _onBoxUpdated;
    private _sceneUpdated;
    private maxMemory;
    /**
     * Indicates whether the scene needs to be re-rendered.
     * Can only be set to true. Cleared on each render.
     */
    get needsUpdate(): boolean;
    set needsUpdate(value: boolean);
    /**
   * Indicates whether the next render should skip antialiasing.
   * Useful for expensive operations such as the section box.
   * Can only be set to true. Cleared on each render.
   */
    get skipAntialias(): boolean;
    set skipAntialias(value: boolean);
    constructor(scene: RenderScene, viewport: Viewport, materials: ViewerMaterials, camera: Camera, settings: Settings);
    /**
     * Removes all objects from rendering and disposes the WebGL context.
     */
    dispose(): void;
    /**
     * Gets or sets the background color or texture of the scene.
     */
    get background(): THREE.Color | THREE.Texture;
    set background(color: THREE.Color | THREE.Texture);
    /**
     * Signal dispatched at the end of each frame if the scene was updated, such as visibility changes.
     */
    get onSceneUpdated(): import("ste-signals").ISignal;
    /**
     * Signal dispatched when bounding box is updated.
     */
    get onBoxUpdated(): import("ste-signals").ISignal;
    /**
     * Determines whether text rendering is enabled or not.
     */
    get textEnabled(): boolean;
    set textEnabled(value: boolean);
    /**
     * Returns the bounding box encompassing all rendered objects.
     * @param target - Box in which to copy the result. A new instance is created if undefined.
     * @returns The bounding box encompassing all rendered objects.
     */
    getBoundingBox(target?: THREE.Box3): THREE.Box3;
    /**
     * Updates the global rendering bounding box.
     * @param box - The new bounding box.
     */
    updateBox(box: THREE.Box3): void;
    /**
     * Notifies that a scene was updated this frame.
     */
    notifySceneUpdate(): void;
    /**
     * Renders what is in the camera's view.
     */
    render(): void;
    /**
     * Adds an object to be rendered.
     * @param target The object or scene to add for rendering.
     */
    add(target: Scene | THREE.Object3D): boolean;
    /**
     * Removes an object from rendering.
     * @param target The object or scene to remove from rendering.
     */
    remove(target: Scene | THREE.Object3D): void;
    /**
     * Clears all objects from rendering.
     */
    clear(): void;
    /**
     * Returns an estimate of the memory used by the renderer.
     */
    get estimatedMemory(): number;
    /**
     * Determines the target sample count on the rendering target.
     * Higher number increases quality.
     */
    get samples(): number;
    set samples(value: number);
    private fitViewport;
}
